#define ATMEGA32U4 1//* put here the uC you are using, like in the lines above followed by "1", like "ATMEGA328 1", "DEBUG 1", etc.
#define USING_MUX 1 //* comment if not using a multiplexer, uncomment if using it.

#include "MIDIUSB.h"
#include <Multiplexer4067.h> // Multiplexer CD4067 library >> https://github.com/sumotoy/Multiplexer4067
#include <Thread.h> // Threads library >> https://github.com/ivanseidel/ArduinoThread
#include <ThreadController.h> // Same as above

#define N_MUX 3 //* number of multiplexers
#define s0 2
#define s1 4
#define s2 5
#define s3 6
#define x1 A0 // analog pin of the first mux
#define x2 A1 // analog pin of the second mux...
#define x3 A2 // digital pin of the third mux...

// Initializes the multiplexer
Multiplexer4067 mux[N_MUX] = {
  Multiplexer4067(s0, s1, s2, s3, x1),
  Multiplexer4067(s0, s1, s2, s3, x2),
  Multiplexer4067(s0, s1, s2, s3, x3)
};

const int N_BUTTONS = 5 + 6 + 0 + 9; //*  total numbers of buttons. Number of buttons in the Arduino + number of buttons on multiplexer 1 + number of buttons on multiplexer 2...
const int N_BUTTONS_ARDUINO = 5; //* number of buttons connected straight to the Arduino (in order)
const int BUTTON_ARDUINO_PIN[N_BUTTONS] = {7, 8, 9, 10, 16}; //* pins of each button connected straight to the Arduino


#define USING_BUTTON_CC_N 1
int BUTTON_CC_N[N_BUTTONS] = {29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48}; //* Add the CC NUMBER of each button/switch you want

const int N_BUTTONS_PER_MUX[N_MUX] = {6, 0, 9}; //* number of buttons in each mux (in order)
const int BUTTON_MUX_PIN[N_MUX][16] = { //* pin of each button of each mux in order
    {10, 11, 12, 13, 14, 15}, //* pins of the first mux
    {}, //* pins of the second
    {1, 2, 3, 4, 5, 6, 7, 8, 9} //* pins of the third
};

int buttonMuxThreshold = 500;

int buttonCState[N_BUTTONS] = {0};        // stores the button current value
int buttonPState[N_BUTTONS] = {0};        // stores the button previous value

// debounce
unsigned long lastDebounceTime[N_BUTTONS] = {0};  // the last time the output pin was toggled
unsigned long debounceDelay = 5;    //* the debounce time; increase if the output flickers

// velocity
byte velocity[N_BUTTONS] = {127};

// POTENTIOMETERS
const int N_POTS = 0 + 10 + 16 + 2; //* total numbers of pots (slide & rotary). Number of pots in the Arduino + number of pots on multiplexer 1 + number of pots on multiplexer 2...
const int N_POTS_ARDUINO = 0; //* number of pots connected straight to the Arduino
const int POT_ARDUINO_PIN[N_POTS_ARDUINO] = {}; //* pins of each pot connected straight to the Arduino

int POT_CC_N[N_POTS] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28}; // Add the CC NUMBER of each pot you want

const int N_POTS_PER_MUX[N_MUX] = {10, 16, 2}; //* number of pots in each multiplexer (in order)
const int POT_MUX_PIN[N_MUX][16] = { //* pins of each pot of each mux in the order you want them to be
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10}, // pins of the first mux
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}, // pins of the second mux
    {0, 10} // pins of the third mux
};

int potCState[N_POTS] = {0}; // Current state of the pot
int potPState[N_POTS] = {0}; // Previous state of the pot
int potVar = 0; // Difference between the current and previous state of the pot

int potMidiCState[N_POTS] = {0}; // Current state of the midi value
int potMidiPState[N_POTS] = {0}; // Previous state of the midi value

const int TIMEOUT = 300; //* Amount of time the potentiometer will be read after it exceeds the varThreshold
const int varThreshold = 10; //* Threshold for the potentiometer signal variation
boolean potMoving = true; // If the potentiometer is moving
unsigned long PTime[N_POTS] = {0}; // Previously stored time
unsigned long timer[N_POTS] = {0}; // Stores the time that has elapsed since the timer was reset

byte MIDI_CH = 1; //* MIDI channel to be used
byte NOTE = 36; //* Lowest NOTE to be used - if not using custom NOTE NUMBER
byte CC = 1; //* Lowest MIDI CC to be used - if not using custom CC NUMBER

// LED
byte ledPin = 3;

void setup() {
    Serial.begin(115200);

    // Initialize buttons with pull up resistors
    for (int i = 0; i < N_BUTTONS_ARDUINO; i++) {
        pinMode(BUTTON_ARDUINO_PIN[i], INPUT_PULLUP);
    }

    // Initialize the multiplexers
    for (int i = 0; i < N_MUX; i++) {
        mux[i].begin();
    }
    //* Set each X pin as input_pullup (avoid floating behavior)
    pinMode(x1, INPUT_PULLUP);
    pinMode(x2, INPUT_PULLUP);

    // Threads
    threadPotentiometers.setInterval(10); // every how many millisiconds
    threadPotentiometers.onRun(potentiometers); // the function that will be added to the thread
    cpu.add(&threadPotentiometers); // add every thread here

    // LED
    pinMode(ledPin, OUTPUT);
    analogWrite(ledPin, 100);
}

void loop() {
    cpu.run(); // for threads
    buttons();
}

// BUTTONS
void buttons() {
    // read pins from arduino
    for (int i = 0; i < N_BUTTONS_ARDUINO; i++) {
        buttonCState[i] = !(digitalRead(BUTTON_ARDUINO_PIN[i]));
    }

    int nButtonsPerMuxSum = N_BUTTONS_ARDUINO; // offsets the buttonCState at every mux reading
    // read the pins from every mux
    for (int j = 0; j < N_MUX; j++) {
        for (int i = 0; i < N_BUTTONS_PER_MUX[j]; i++) {
            buttonCState[i + nButtonsPerMuxSum] = mux[j].readChannel(BUTTON_MUX_PIN[j][i]);
            // Scale values to 0-1
            if (buttonCState[i + nButtonsPerMuxSum] > buttonMuxThreshold) {
                buttonCState[i + nButtonsPerMuxSum] = LOW;
            } else {
                buttonCState[i + nButtonsPerMuxSum] = HIGH;
            }
        }
        nButtonsPerMuxSum += N_BUTTONS_PER_MUX[j];
    }

    for (int i = 0; i < N_BUTTONS; i++) {
        if ((millis() - lastDebounceTime[i]) > debounceDelay) {
            if (buttonPState[i] != buttonCState[i]) {
                lastDebounceTime[i] = millis();
                if (buttonCState[i] == LOW) {
                    velocity[i] = 127; // if button is pressed velocity is 127
                } else {
                    velocity[i] = 0; // if button is released velocity is 0
                }
                controlChange(MIDI_CH, BUTTON_CC_N[i], velocity[i]); //  (channel, CC number,  CC value)
                MidiUSB.flush();
                buttonPState[i] = buttonCState[i];
            }
        }
    }
}

  /////////////////////////////////////////////
  // POTENTIOMETERS
  void potentiometers() {
    // reads the pins from arduino
    for (int i = 0; i < N_POTS_ARDUINO; i++) {
        potCState[i] = analogRead(POT_ARDUINO_PIN[i]);
    }

    int nPotsPerMuxSum = N_POTS_ARDUINO; //offsets the buttonCState at every mux reading

    // reads the pins from every mux
    for (int j = 0; j < N_MUX; j++) {
        for (int i = 0; i < N_POTS_PER_MUX[j]; i++) {
            potCState[i + nPotsPerMuxSum] = mux[j].readChannel(POT_MUX_PIN[j][i]);
        }
        nPotsPerMuxSum += N_POTS_PER_MUX[j];
    }

    for (int i = 0; i < N_POTS; i++) { // Loops through all the potentiometers
        potMidiCState[i] = map(potCState[i], 0, 1023, 0, 127); // Maps the reading of the potCState to a value usable in midi
        potVar = abs(potCState[i] - potPState[i]); // Calculates the absolute value between the difference between the current and previous state of the pot
        if (potVar > varThreshold) { // Opens the gate if the potentiometer variation is greater than the threshold
            PTime[i] = millis(); // Stores the previous time
        }

        timer[i] = millis() - PTime[i]; // Resets the timer 11000 - 11000 = 0ms

        if (timer[i] < TIMEOUT) { // If the timer is less than the maximum allowed time it means that the potentiometer is still moving
            potMoving = true;
        } else {
            potMoving = false;
        }

        if (potMoving == true) { // If the potentiometer is still moving, send the change control
            if (potMidiPState[i] != potMidiCState[i]) {
                controlChange(MIDI_CH, POT_CC_N[i], potMidiCState[i]); //  (channel, CC number,  CC value)
                MidiUSB.flush();
                potPState[i] = potCState[i]; // Stores the current reading of the potentiometer to compare with the next
                potMidiPState[i] = potMidiCState[i];
            }
        }
    }
}

void controlChange(byte channel, byte control, byte value) {
    midiEventPacket_t event = {0x0B, 0xB0 | channel, control, value};
    MidiUSB.sendMIDI(event);
}
